from .operations import apply_phase
from .operations import apply_operator
from .operations import apply_to_qubits
from .operations import apply_all_qubits
from .operations import act_last_qubit
from .operations import swap
from .operations import multi_cz
from .operations import multi_controlled_x, multi_controlled_z
from .operations import act_first_qubits
from .measurement import measure

